#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>
#include <string.h>
#include <sys/mman.h>
#include <fcntl.h>
#include <getopt.h>
#include <errno.h>

#include "queue.h"
#include "vuart.h"

#define MAX_BUFSIZE 2048

static struct option long_options[] = {
	{"domain", 1, 0, 'd'},
	{"uart", 1, 0, 'u'},
	{"init", 0, 0, 'i'},
	{"test", 0, 0, 't'},
	{"start", 0, 0, 's'},
	{"niter", 1, 0, 'n'},
	{"help", 0, 0, 'h'},
	{0, 0, 0, 0}
};

const char *usage =
"termios --domain <domain> --uart <uart id> --init --start --niter <iterations> --help\n"
"Usage options:\n"
"-h, --help             : Print this help\n";

void print_usage(int exit_code)
{
	printf(usage);
	exit(exit_code);
}

static unsigned long get_uart_memaddr(int uart)
{
	if (uart == 0)
		return (unsigned long)(496 << 20);
	else
		return (unsigned long)(500 << 20);
}

static size_t get_uart_memsize(void)
{
	return (4 << 20);
}

static int get_uart_count(void)
{
	return 32;
}

static void *map_vuart_memory(int dev_id)
{
	int fd = -1;
	void *mmap_vaddr = 0;
	size_t size = get_uart_memsize();
	off_t pbase = get_uart_memaddr(dev_id);

        fd = open("/dev/mem", O_RDWR);
	if (fd < 0) {
                printf("Cannot open /dev/mem: \"%s\"\n", strerror(errno));
                exit(-1);
        }

	mmap_vaddr = mmap(NULL, get_uart_memsize(), PROT_READ | PROT_WRITE, MAP_SHARED, fd, pbase);
	if (mmap_vaddr == MAP_FAILED) {
                printf("mmap error (%s)\n", strerror(errno));
		exit(-1);
        }
	{
		volatile unsigned int *ptr = mmap_vaddr;
		ptr[0] = 0xdeadbeef;
		printf("mmap[0] = %x, mmap[1] = %x\n", ptr[0], ptr[1]);
	}
	printf("mmapped  virt %p size 0x%x pbase=%lx\n", mmap_vaddr, size, pbase);

	return mmap_vaddr;
}

static void putstr(const char *s, int vuart, int qid)
{
	for (; *s != 0; s++)
		outbyte(*s, vuart, qid);
}

static size_t getstr(char *buf, size_t bufsize, int vuart, int qid)
{
	int i = 0;
	int c;

	for (;;) {
		if (i == bufsize - 1) break;

		c = inbyte_nowait(vuart, qid);
		if (c == EOF) break;

		buf[i] = c;
		i++;
	}

	buf[i] = 0;
	return i;
}

int main(int argc, char** argv)
{
	char  buf[MAX_BUFSIZE];
	char c;

	int domain = 0, uart = 0;
	int initq = 0;
	int test_uart = 0;
	int start = 0;
	unsigned long long niter = ~0;

	int opt_idx = 0;

        void *vuart_baseaddr = 0;

	while((c = getopt_long(argc, argv, "histd:u:n:", long_options, &opt_idx)) >= 0) {
		switch(c) {
		case 'i':
			initq = 1;
			break;
		case 't':
			test_uart = 1;
			break;
		case 's':
			start = 1;
			break;
		case 'd':
			domain = (int)strtoul(optarg, 0, 10);
			if( domain < 0) {
				printf("negative domain id (%d) specified?\n", domain);
				exit(-1);
			}
			break;
		case 'n':
			niter = (int)strtoull(optarg, 0, 10);
			if (niter < 0) {
				printf("negative niter (%d) specified?\n", niter);
				exit(-1);
			}
			break;
		case 'u':
			uart = (int)strtoul(optarg, 0, 10);
			if( uart != 0 && uart != 1) {
				printf("Wrong uart id (%d) specified\n", uart);
				exit(-1);
			}
			break;
		case 'h':
			print_usage(0);
			break;
		default:
			break;
		}
	}

	printf("uart: %d domain: %d\n", uart, domain);

	vuart_baseaddr = map_vuart_memory(uart);

	uart_init(vuart_baseaddr, 8, initq);

	if (test_uart) {
		printf("Testing domain queues...\n", domain);

		putstr("Hello, World!", domain, DOMU_WRITE_Q);
		getstr(buf, MAX_BUFSIZE, domain, DOM0_READ_Q);
		printf("[%s]\n", buf);
	}

	if (start) {
		unsigned long long i = 0;

		printf("=================================================\n");
		for (i = 0; i < niter; i++) {
			size_t size = getstr(buf, MAX_BUFSIZE, domain, DOM0_READ_Q);
			if (size) printf("%s", buf);
		}
	}

	return 0;
}
